//
// Moravec Interest Point Detector
//

#include	<stdlib.h>
#include	<math.h>
#include	"Moravec.h"
#include	"Tools.h"

#define		MaxGray		255

// Moravec operator w/o output image
int
Moravec(int *Im, int width, int height,
	int **up0, int **vp0, int *cnr_num,
	double ratio, int max_corner, double min_dist)
{
   int		mo, d;
   int		corner_num;
   int		x, u0, v0;
   int		width1, height1;

   double	mo_max, dd, mo_thd;
   
   int		*up = new int [max_corner];
   int		*vp = new int [max_corner];

   float	*tmp = new float [width*height];
   
   double	*dp = new double [max_corner];

   (*up0) = up;
   (*vp0) = vp;

   width1 = width - 1;
   height1 = height - 1;

   // clear image boundary
   for (x = 0; x < height; x++)
      tmp[x] = tmp[width * height1 + x] = 0.0;

   for (x = 1; x < height1; x++)
      tmp[width * x] = tmp[width * x + width1] = 0.0;
   
   // calc moverac
   mo_max = 0.0;
   for (v0 = 1; v0 < height1; v0++)
   {
      x = (v0 * width) + 1;
      for (u0 = 1; u0 < width1; u0++, x++)
      {
	 d = Im[x];
	 mo = 0;
	 for (int k = -1; k <= 1; k++)
	 {
	    mo += abs(Im[x + k - width] - d)
	       + abs(Im[x + k] - d)
	       + abs(Im[x + k + width] - d);
	 }
	 tmp[x] = mo;

	 if (mo_max < (double)mo)
	    mo_max = (double)mo;
      }
   }

   // threshold for moverac
   mo_thd = mo_max * ratio;

   // check local minima
   corner_num = 0;
   for (v0 = 1; v0 < height1; v0++)
   {
      x = v0 * width + 1;
      for (u0 = 1; u0 < width1; u0++, x++)
      {
	 dd = tmp[x];
	 if (dd < mo_thd)  continue;

	 if (   check_ge_upperleft(dd, tmp, x, width)
             && check_gt_lowerright(dd, tmp, x, width))
	 {
	    corner_num = add_list(u0, v0, dd, up, vp, dp,
				  corner_num, max_corner, min_dist);
	 }
      }
   }

// delete arrays
   delete [] tmp;
   
   return (*cnr_num = corner_num);
}

// Moravec operator with output image
int
Moravec(int *Im, int *Imo, int width, int height,
	int **up0, int **vp0, int *cnr_num,
	double ratio, int max_corner, double min_dist)
{
   int		x, corner_num;

   // feature detection
   corner_num = Moravec(Im, width, height, up0, vp0, cnr_num,
			ratio, max_corner, min_dist);

   // copy image
   x = 0;
   for (int v = 0; v < height; v++)
      for (int u = 0; u < width; u++)
	 Imo[x++] = Im[x];

   // plot corners
   for (int i = 0; i < corner_num; i++)
   {
      x = (*vp0)[i] * width + (*up0)[i];
      Imo[x - width - 1] = MaxGray;
      Imo[x - width]     = MaxGray;
      Imo[x - width + 1] = MaxGray;
      Imo[x - 1]         = MaxGray;
      Imo[x]             = 0;
      Imo[x + 1]         = MaxGray;
      Imo[x + width - 1] = MaxGray;
      Imo[x + width]     = MaxGray;
      Imo[x + width + 1] = MaxGray;
   }

   return corner_num;
}
